from __future__ import annotations

import logging
from dataclasses import dataclass, field
from enum import IntEnum, Enum
from functools import reduce
from operator import or_

from krrood.entity_query_language.entity import (
    symbolic_mode,
    let,
    an,
    entity,
    not_,
    in_,
    From,
    for_all,
)
from numpy import ndarray
from probabilistic_model.probabilistic_circuit.rx.helper import (
    uniform_measure_of_simple_event,
)
from random_events.interval import Bound
from random_events.product_algebra import *
from typing_extensions import Generic, TypeVar

from ..datastructures.prefixed_name import PrefixedName
from ..datastructures.variables import SpatialVariables
from ..spatial_types.derivatives import DerivativeMap
from ..spatial_types.spatial_types import (
    TransformationMatrix,
    Vector3,
    Point3,
)
from ..utils import IDGenerator
from ..semantic_annotations.semantic_annotations import (
    Container,
    Handle,
    Dresser,
    Drawer,
    Door,
    Wall,
    DoubleDoor,
    Room,
    Floor,
)
from ..world import World
from ..world_description.connections import (
    PrismaticConnection,
    FixedConnection,
    RevoluteConnection,
)
from ..world_description.degree_of_freedom import DegreeOfFreedom
from ..world_description.geometry import Scale
from ..world_description.shape_collection import BoundingBoxCollection, ShapeCollection
from ..world_description.world_entity import Body, Region, CollisionCheckingConfig

id_generator = IDGenerator()


class Direction(IntEnum):
    X = 0
    Y = 1
    Z = 2
    NEGATIVE_X = 3
    NEGATIVE_Y = 4
    NEGATIVE_Z = 5


@dataclass
class HasDoorLikeFactories(ABC):
    """
    Mixin for factories receiving multiple DoorLikeFactories.
    """

    door_factories: List[DoorLikeFactory] = field(default_factory=list, hash=False)
    """
    The factories used to create the doors.
    """

    door_transforms: List[TransformationMatrix] = field(
        default_factory=list, hash=False
    )
    """
    The transformations for the doors relative their parent container.
    """

    def _create_door_upper_lower_limits(
        self, semantic_door_annotation: Door
    ) -> Tuple[DerivativeMap[float], DerivativeMap[float]]:
        """
        Return the upper and lower limits for the door's degree of freedom.
        """
        lower_limits = DerivativeMap[float]()
        upper_limits = DerivativeMap[float]()
        lower_limits.position = -np.pi / 2
        upper_limits.position = 0.0

        parent_connection = semantic_door_annotation.handle.body.parent_connection
        door_P_handle: ndarray[float] = (
            parent_connection.origin_expression.to_position().to_np()
        )
        if np.sign(door_P_handle[1]) < 0:
            lower_limits.position = 0.0
            upper_limits.position = np.pi / 2
        return upper_limits, lower_limits

    def _add_hinge_to_door(
        self, door_factory: DoorFactory, parent_T_door: TransformationMatrix
    ):
        """
        Adds a hinge to the door. The hinge's pivot point is on the opposite side of the handle.
        :param door_factory: The factory used to create the door.
        :param parent_T_door: The transformation matrix defining the door's position and orientation relative
        """
        door_world = door_factory.create()
        root = door_world.root
        semantic_door_annotation: Door = door_world.get_semantic_annotations_by_type(
            Door
        )[0]

        door_hinge = Body(
            name=PrefixedName(f"{root.name.name}_door_hinge", root.name.prefix)
        )
        parent_T_hinge = self._calculate_door_pivot_point(
            semantic_door_annotation, parent_T_door, door_factory.scale
        )
        hinge_T_door = parent_T_hinge.inverse() @ parent_T_door

        hinge_door_connection = FixedConnection(
            parent=door_hinge, child=root, parent_T_connection_expression=hinge_T_door
        )
        with door_world.modify_world():
            door_world.add_connection(hinge_door_connection)

        return door_world, parent_T_hinge

    def _add_door_to_world(
        self,
        door_factory: DoorFactory,
        parent_T_door: TransformationMatrix,
        parent_world: World,
    ):
        """
        Adds a door to the parent world using a new door hinge body with a revolute connection.

        :param door_factory: The factory used to create the door.
        :param parent_T_door: The transformation matrix defining the door's position and orientation relative
        to the parent world.
        :param parent_world: The world to which the door will be added.
        """
        with parent_world.modify_world():
            door_world, parent_T_hinge = self._add_hinge_to_door(
                door_factory, parent_T_door
            )
            semantic_door_annotation: Door = (
                door_world.get_semantic_annotations_by_type(Door)[0]
            )
            upper_limits, lower_limits = self._create_door_upper_lower_limits(
                semantic_door_annotation
            )

            root = door_world.root
            dof = DegreeOfFreedom(
                name=PrefixedName(f"{root.name.name}_connection", root.name.prefix),
                lower_limits=lower_limits,
                upper_limits=upper_limits,
            )
            with parent_world.modify_world():
                parent_world.add_degree_of_freedom(dof)
                connection = RevoluteConnection(
                    parent=parent_world.root,
                    child=root,
                    parent_T_connection_expression=parent_T_hinge,
                    multiplier=1.0,
                    offset=0.0,
                    axis=Vector3.Z(),
                    dof_name=dof.name,
                )

                parent_world.merge_world(door_world, connection)

    def add_doorlike_semantic_annotation_to_world(
        self,
        parent_world: World,
    ):
        """
        Adds door-like semantic annotations to the parent world.
        """
        for door_factory, door_transform in zip(
            self.door_factories, self.door_transforms
        ):
            if isinstance(door_factory, DoorFactory):
                self._add_door_to_world(
                    door_factory=door_factory,
                    parent_T_door=door_transform,
                    parent_world=parent_world,
                )
            elif isinstance(door_factory, DoubleDoorFactory):
                self._add_double_door_to_world(
                    door_factory=door_factory,
                    parent_T_door=door_transform,
                    parent_world=parent_world,
                )

    def _calculate_door_pivot_point(
        self,
        semantic_door_annotation: Door,
        parent_T_door: TransformationMatrix,
        scale: Scale,
    ) -> TransformationMatrix:
        """
        Calculate the door pivot point based on the handle position and the door scale. The pivot point is on the opposite
        side of the handle.

        :param semantic_door_annotation: The door semantic annotation containing the handle.
        :param parent_T_door: The transformation matrix defining the door's position and orientation.
        :param scale: The scale of the door.

        :return: The transformation matrix defining the door's pivot point.
        """
        connection = semantic_door_annotation.handle.body.parent_connection
        door_P_handle: ndarray[float] = (
            connection.origin_expression.to_position().to_np()
        )

        sign = np.sign(door_P_handle[1]) if door_P_handle[1] != 0 else 1
        offset = sign * (scale.y / 2)
        parent_P_hinge = parent_T_door.to_np()[:3, 3] + np.array([0, offset, 0])

        parent_T_hinge = TransformationMatrix.from_point_rotation_matrix(
            Point3(*parent_P_hinge)
        )

        return parent_T_hinge

    def _add_double_door_to_world(
        self,
        door_factory: DoubleDoorFactory,
        parent_T_door: TransformationMatrix,
        parent_world: World,
    ):
        """
        Adds a double door to the parent world with a fixed connection.
        :param door_factory: The factory used to create the double door.
        :param parent_T_door: The transformation matrix defining the double door's position and orientation relative
        to the parent world.
        :param parent_world: The world to which the double door will be added.
        """
        door_world = door_factory.create()
        connection = FixedConnection(
            parent=parent_world.root,
            child=door_world.root,
            parent_T_connection_expression=parent_T_door,
        )

        parent_world.merge_world(door_world, connection)

    def remove_doors_from_world(
        self, parent_world: World, wall_event_thickness: float = 0.1
    ):
        """
        Remove the door volumes from all bodies in the world that are not doors.

        :param parent_world: The world from which to remove the door volumes.
        :param wall_event_thickness: The thickness of the wall event used to create the door events.
        """
        doors: List[Door] = parent_world.get_semantic_annotations_by_type(Door)
        if not doors:
            return
        all_doors_event = self._build_all_doors_event_from_semantic_annotations(
            doors, wall_event_thickness
        )

        all_bodies_not_door = self._get_all_bodies_excluding_doors_from_world(
            parent_world
        )

        if not all_doors_event.is_empty():
            self._remove_doors_from_bodies(all_bodies_not_door, all_doors_event)

    def _get_all_bodies_excluding_doors_from_world(self, world: World) -> List[Body]:
        """
        Return all bodies in the world that are not part of any door semantic annotation.

        :param world: The world from which to get the bodies.
        :return: A list of bodies that are not part of any door semantic annotation.
        """
        with symbolic_mode():
            all_doors = Door(From(world.semantic_annotations))
            other_body = let(type_=Body, domain=world.bodies)
            door_bodies = all_doors.bodies
            bodies_without_excluded_bodies_query = an(
                entity(
                    other_body, for_all(door_bodies, not_(in_(other_body, door_bodies)))
                )
            )

        filtered_bodies = list(bodies_without_excluded_bodies_query.evaluate())
        return filtered_bodies

    def _build_all_doors_event_from_semantic_annotations(
        self, doors: List[Door], wall_event_thickness: float = 0.1
    ) -> Event:
        """
        Build a single event representing all doors by combining the events of each door.

        :param doors: The list of door semantic annotations to build the event from.
        :param wall_event_thickness: The thickness of the wall event used to create the door events.
        :return: An event representing all doors.
        """
        door_events = [
            self._build_single_door_event(door, wall_event_thickness) for door in doors
        ]
        if door_events:
            return reduce(or_, door_events)
        return Event()

    def _build_single_door_event(
        self, door: Door, wall_event_thickness: float = 0.1
    ) -> Event:
        """
        Build an event representing a single door by creating a bounding box event around the door's collision shapes

        :param door: The door semantic annotation to build the event from.
        :param wall_event_thickness: The thickness of the wall event used to create the door event.
        :return: An event representing the door.
        """
        door_event = door.body.collision.as_bounding_box_collection_in_frame(
            door._world.root
        ).event

        door_plane_spatial_variables = SpatialVariables.yz
        door_thickness_spatial_variable = SpatialVariables.x.value
        door_event = door_event.marginal(door_plane_spatial_variables)
        door_event.fill_missing_variables([door_thickness_spatial_variable])
        thickness_event = SimpleEvent(
            {
                door_thickness_spatial_variable: closed(
                    -wall_event_thickness / 2, wall_event_thickness / 2
                )
            }
        ).as_composite_set()
        thickness_event.fill_missing_variables(door_plane_spatial_variables)
        door_event = door_event & thickness_event

        return door_event

    def _remove_doors_from_bodies(self, bodies: List[Body], all_doors_event: Event):
        """
        Remove the door volumes from the given bodies by subtracting the all_doors_event from each body's collision event.

        :param bodies: The list of bodies from which to remove the door volumes.
        :param all_doors_event: The event representing all doors.
        """
        for body in bodies:
            self._remove_door_from_body(body, all_doors_event)

    def _remove_door_from_body(self, body: Body, all_doors_event: Event):
        """
        Remove the door volumes from the given body by subtracting the all_doors_event from the body's collision event.

        :param body: The body from which to remove the door volumes.
        :param all_doors_event: The event representing all doors.
        """
        root = body._world.root
        body_event = (
            body.collision.as_bounding_box_collection_in_frame(root).event
            - all_doors_event
        )
        new_collision = BoundingBoxCollection.from_event(root, body_event).as_shapes()
        body.collision = new_collision
        body.visual = new_collision


class IntervalConstants:
    """
    Predefined intervals for semantic directions.
    """

    ZERO_DIRAC = (0, 0, Bound.CLOSED, Bound.CLOSED)
    ZERO_TO_ONE_THIRD = (0, 1 / 3, Bound.CLOSED, Bound.CLOSED)
    ONE_THIRD_TO_TWO_THIRD = (1 / 3, 2 / 3, Bound.OPEN, Bound.OPEN)
    HALF_DIRAC = (0.5, 0.5, Bound.CLOSED, Bound.CLOSED)
    TWO_THIRD_TO_ONE = (2 / 3, 1, Bound.CLOSED, Bound.CLOSED)
    ONE_DIRAC = (1, 1, Bound.CLOSED, Bound.CLOSED)


class SemanticDirection(Enum): ...


class HorizontalSemanticDirection(SimpleInterval, SemanticDirection):
    """
    Semantic directions for horizontal positioning.
    """

    FULLY_LEFT = IntervalConstants.ZERO_DIRAC
    LEFT = IntervalConstants.ZERO_TO_ONE_THIRD
    CENTER = IntervalConstants.ONE_THIRD_TO_TWO_THIRD
    FULLY_CENTER = IntervalConstants.HALF_DIRAC
    RIGHT = IntervalConstants.TWO_THIRD_TO_ONE
    FULLY_RIGHT = IntervalConstants.ONE_DIRAC


class VerticalSemanticDirection(SimpleInterval, SemanticDirection):
    """
    Semantic directions for vertical positioning.
    """

    FULLY_TOP = IntervalConstants.ZERO_DIRAC
    TOP = IntervalConstants.ZERO_TO_ONE_THIRD
    CENTER = IntervalConstants.ONE_THIRD_TO_TWO_THIRD
    FULLY_CENTER = IntervalConstants.HALF_DIRAC
    BOTTOM = IntervalConstants.TWO_THIRD_TO_ONE
    FULLY_BOTTOM = IntervalConstants.ONE_DIRAC


@dataclass
class SemanticPositionDescription:
    """
    Describes a position by mapping semantic concepts (RIGHT, CENTER, LEFT, TOP, BOTTOM) to instances of
    random_events.intervals.SimpleInterval, which are then used to "zoom" into specific regions of an event.
    Each DirectionInterval divides the original event into three parts, either vertically or horizontally, and zooms into
    one of them depending on which specific direction was chosen.
    The sequence of zooms is defined by the order of directions in the horizontal_direction_chain and
    vertical_direction_chain lists.
    Finally, we can sample aa 2d pose from the resulting event
    """

    horizontal_direction_chain: List[HorizontalSemanticDirection]
    """
    Describes the sequence of zooms in the horizontal direction (Y axis).
    """

    vertical_direction_chain: List[VerticalSemanticDirection]
    """
    Describes the sequence of zooms in the vertical direction (Z axis).
    """

    @staticmethod
    def _zoom_interval(base: SimpleInterval, target: SimpleInterval) -> SimpleInterval:
        """
        Zoom 'base' interval by the percentage interval 'target' (0..1),
        preserving the base's boundary styles.

        :param base: The base interval to be zoomed in.
        :param target: The target interval defining the zoom percentage (0..1).
        :return: A new SimpleInterval representing the zoomed-in interval.
        """
        span = base.upper - base.lower
        new_lower = base.lower + span * target.lower
        new_upper = base.lower + span * target.upper
        return SimpleInterval(new_lower, new_upper, base.left, base.right)

    def _apply_zoom(self, simple_event: SimpleEvent) -> SimpleEvent:
        """
        Apply zooms in order and return the resulting intervals.

        :param simple_event: The event to zoom in.
        :return: A SimpleEvent containing the resulting intervals after applying all zooms.
        """
        simple_events = [
            self._apply_zoom_in_one_direction(
                axis,
                assignment.simple_sets[0],
            )
            for axis, assignment in simple_event.items()
        ]

        if not simple_events:
            return SimpleEvent()

        return reduce(or_, simple_events)

    def _apply_zoom_in_one_direction(
        self, axis: Continuous, current_interval: SimpleInterval
    ) -> SimpleEvent:
        """
        Apply zooms in one direction (Y, horizontal or Z, vertical) in order and return the resulting interval.

        :param axis: The axis to zoom in (SpatialVariables.y or SpatialVariables.z).
        :param current_interval: The current interval to zoom in.
        :return: A SimpleEvent containing the resulting interval after applying all zooms in the specified direction.
        """
        if axis == SpatialVariables.y.value:
            directions = self.horizontal_direction_chain
        elif axis == SpatialVariables.z.value:
            directions = self.vertical_direction_chain
        else:
            raise NotImplementedError

        for step in directions:
            current_interval = self._zoom_interval(current_interval, step)

        return SimpleEvent({axis: current_interval})

    def sample_point_from_event(self, event: Event) -> Tuple[float, float]:
        """
        Sample a 2D point from the given event by applying the zooms defined in the semantic position description.

        :param event: The event to sample from.
        :return: A sampled 2D point as a tuple (y, z).
        """
        simple_event = self._apply_zoom(event.bounding_box())
        event_circuit = uniform_measure_of_simple_event(simple_event)
        return event_circuit.sample(amount=1)[0]


@dataclass
class HasHandleFactory(ABC):
    """
    Mixin for factories receiving a HandleFactory.
    If both parent_T_handle and semantic_position are set, parent_T_handle will be
    prioritized.
    """

    handle_factory: HandleFactory = field(kw_only=True)
    """
    The factory used to create the handle of the door.
    """

    semantic_position: Optional[SemanticPositionDescription] = field(default=None)
    """
    The direction on the door in which the handle positioned.
    """

    parent_T_handle: Optional[TransformationMatrix] = field(default=None)
    """
    The transformation matrix of the handle from the parent
    """

    def __post_init__(self):
        assert (
            self.parent_T_handle is not None or self.semantic_position is not None
        ), "Either parent_T_handle or semantic_position must be set."
        if self.parent_T_handle is not None and self.semantic_position is not None:
            logging.warning(
                f"During the creation of a factory, both parent_T_handle and semantic_position were set. Prioritizing parent_T_handle."
            )

    def create_parent_T_handle_from_parent_scale(
        self, scale: Scale
    ) -> Optional[TransformationMatrix]:
        """
        Return a transformation matrix that defines the position and orientation of the handle relative to its parent.
        :raises: NotImplementedError if the handle direction is Z or NEGATIVE_Z.
        """
        assert self.semantic_position is not None
        sampled_2d_point = self.semantic_position.sample_point_from_event(
            scale.simple_event.as_composite_set().marginal(SpatialVariables.yz)
        )

        return TransformationMatrix.from_xyz_rpy(
            x=scale.x / 2, y=sampled_2d_point[0], z=sampled_2d_point[1]
        )

    def add_handle_to_world(
        self,
        parent_T_handle: TransformationMatrix,
        parent_world: World,
    ):
        """
        Adds a handle to the parent world with a fixed connection.

        :param parent_T_handle: The transformation matrix defining the handle's position and orientation relative
        to the parent world.
        :param parent_world: The world to which the handle will be added.
        """
        handle_world = self.handle_factory.create()

        connection = FixedConnection(
            parent=parent_world.root,
            child=handle_world.root,
            parent_T_connection_expression=parent_T_handle,
        )

        parent_world.merge_world(handle_world, connection)


@dataclass
class HasDrawerFactories(ABC):
    """
    Mixin for factories receiving multiple DrawerFactories.
    """

    drawers_factories: List[DrawerFactory] = field(default_factory=list, hash=False)
    """
    The factories used to create drawers.
    """

    parent_T_drawers: List[TransformationMatrix] = field(
        default_factory=list, hash=False
    )
    """
    The transformations for the drawers their parent container.
    """

    def _add_drawer_to_world(
        self,
        drawer_factory: DrawerFactory,
        parent_T_drawer: TransformationMatrix,
        parent_world: World,
    ):

        lower_limits, upper_limits = self._create_drawer_upper_lower_limits(
            drawer_factory
        )
        drawer_world = drawer_factory.create()
        parent_root = parent_world.root
        child_root = drawer_world.root

        parent_T_drawer.reference_frame = parent_root

        dof = DegreeOfFreedom(
            name=PrefixedName(
                f"{child_root.name.name}_connection", child_root.name.prefix
            ),
            lower_limits=lower_limits,
            upper_limits=upper_limits,
        )
        with parent_world.modify_world():
            parent_world.add_degree_of_freedom(dof)
            connection = PrismaticConnection(
                parent=parent_root,
                child=child_root,
                parent_T_connection_expression=parent_T_drawer,
                multiplier=1.0,
                offset=0.0,
                axis=Vector3.X(),
                dof_name=dof.name,
            )

            parent_world.merge_world(drawer_world, connection)

    def add_drawers_to_world(self, parent_world: World):
        """
        Adds drawers to the parent world. A prismatic connection is created for each drawer.
        """

        for drawer_factory, transform in zip(
            self.drawers_factories, self.parent_T_drawers
        ):
            self._add_drawer_to_world(
                drawer_factory=drawer_factory,
                parent_T_drawer=transform,
                parent_world=parent_world,
            )

    @staticmethod
    def _create_drawer_upper_lower_limits(
        drawer_factory: DrawerFactory,
    ) -> Tuple[DerivativeMap[float], DerivativeMap[float]]:
        """
        Return the upper and lower limits for the drawer's degree of freedom.
        """
        lower_limits = DerivativeMap[float]()
        upper_limits = DerivativeMap[float]()
        lower_limits.position = 0.0
        upper_limits.position = drawer_factory.container_factory.scale.x * 0.75

        return upper_limits, lower_limits


T = TypeVar("T")


@dataclass
class SemanticAnnotationFactory(Generic[T], ABC):
    """
    Abstract factory for the creation of worlds containing a single semantic annotation of type T.
    """

    name: PrefixedName = field(kw_only=True)
    """
    The name of the semantic annotation.
    """

    @abstractmethod
    def _create(self, world: World) -> World:
        """
        Create the world containing a semantic annotation of type T.
        Put the custom logic in here.

        :param world: The world to create the semantic annotation in.
        :return: The world.
        """
        raise NotImplementedError()

    def create(self) -> World:
        """
        Create the world containing a semantic annotation of type T.

        :return: The world.
        """
        world = World(name=self.name.name)
        with world.modify_world():
            world = self._create(world)
        return world


@dataclass
class ContainerFactory(SemanticAnnotationFactory[Container]):
    """
    Factory for creating a container with walls of a specified thickness and its opening in direction.
    """

    scale: Scale = field(default_factory=lambda: Scale(1.0, 1.0, 1.0))
    """
    The scale of the container, defining its size in the world.
    """

    wall_thickness: float = 0.05
    """
    The thickness of the walls of the container.
    """

    direction: Direction = field(default=Direction.X)
    """
    The direction in which the container is open.
    """

    def _create(self, world: World) -> World:
        """
        Return a world with a container body at its root.
        """

        container_event = self._create_container_event()

        container_body = Body(name=self.name)
        collision_shapes = BoundingBoxCollection.from_event(
            container_body, container_event
        ).as_shapes()
        container_body.collision = collision_shapes
        container_body.visual = collision_shapes

        semantic_container_annotation = Container(body=container_body, name=self.name)

        world.add_kinematic_structure_entity(container_body)
        world.add_semantic_annotation(semantic_container_annotation)

        return world

    def _create_container_event(self) -> Event:
        """
        Return an event representing a container with walls of a specified thickness.
        """
        outer_box = self.scale.simple_event
        inner_scale = Scale(
            self.scale.x - self.wall_thickness,
            self.scale.y - self.wall_thickness,
            self.scale.z - self.wall_thickness,
        )
        inner_box = inner_scale.simple_event

        inner_box = self._extend_inner_event_in_direction(
            inner_event=inner_box, inner_scale=inner_scale
        )

        container_event = outer_box.as_composite_set() - inner_box.as_composite_set()

        return container_event

    def _extend_inner_event_in_direction(
        self, inner_event: SimpleEvent, inner_scale: Scale
    ) -> SimpleEvent:
        """
        Extend the inner event in the specified direction to create the container opening in that direction.

        :param inner_event: The inner event representing the inner box.
        :param inner_scale: The scale of the inner box used how far to extend the inner event.

        :return: The modified inner event with the specified direction extended.
        """

        match self.direction:
            case Direction.X:
                inner_event[SpatialVariables.x.value] = closed(
                    -inner_scale.x / 2, self.scale.x / 2
                )
            case Direction.Y:
                inner_event[SpatialVariables.y.value] = closed(
                    -inner_scale.y / 2, self.scale.y / 2
                )
            case Direction.Z:
                inner_event[SpatialVariables.z.value] = closed(
                    -inner_scale.z / 2, self.scale.z / 2
                )
            case Direction.NEGATIVE_X:
                inner_event[SpatialVariables.x.value] = closed(
                    -self.scale.x / 2, inner_scale.x / 2
                )
            case Direction.NEGATIVE_Y:
                inner_event[SpatialVariables.y.value] = closed(
                    -self.scale.y / 2, inner_scale.y / 2
                )
            case Direction.NEGATIVE_Z:
                inner_event[SpatialVariables.z.value] = closed(
                    -self.scale.z / 2, inner_scale.z / 2
                )

        return inner_event


@dataclass
class HandleFactory(SemanticAnnotationFactory[Handle]):
    """
    Factory for creating a handle with a specified scale and thickness.
    The handle is represented as a box with an inner cutout to create the handle shape.
    """

    scale: Scale = field(default_factory=lambda: Scale(0.05, 0.1, 0.02))
    """
    The scale of the handle.
    """

    thickness: float = 0.01
    """
    Thickness of the handle bar.
    """

    def _create(self, world: World) -> World:
        """
        Create a world with a handle body at its root.
        """

        handle_event = self._create_handle_event()

        handle = Body(name=self.name)
        collision = BoundingBoxCollection.from_event(handle, handle_event).as_shapes()
        handle.collision = collision
        handle.visual = collision

        semantic_handle_annotation = Handle(name=self.name, body=handle)

        world.add_kinematic_structure_entity(handle)
        world.add_semantic_annotation(semantic_handle_annotation)
        return world

    def _create_handle_event(self) -> Event:
        """
        Return an event representing a handle.
        """

        handle_event = self._create_outer_box_event().as_composite_set()

        inner_box = self._create_inner_box_event().as_composite_set()

        handle_event -= inner_box

        return handle_event

    def _create_outer_box_event(self) -> SimpleEvent:
        """
        Return an event representing the main body of a handle.
        """
        x_interval = closed(0, self.scale.x)
        y_interval = closed(-self.scale.y / 2, self.scale.y / 2)
        z_interval = closed(-self.scale.z / 2, self.scale.z / 2)

        handle_event = SimpleEvent(
            {
                SpatialVariables.x.value: x_interval,
                SpatialVariables.y.value: y_interval,
                SpatialVariables.z.value: z_interval,
            }
        )

        return handle_event

    def _create_inner_box_event(self) -> SimpleEvent:
        """
        Return an event used to cut out the inner part of the handle.
        """
        x_interval = closed(0, self.scale.x - self.thickness)
        y_interval = closed(
            -self.scale.y / 2 + self.thickness, self.scale.y / 2 - self.thickness
        )
        z_interval = closed(-self.scale.z, self.scale.z)

        inner_box = SimpleEvent(
            {
                SpatialVariables.x.value: x_interval,
                SpatialVariables.y.value: y_interval,
                SpatialVariables.z.value: z_interval,
            }
        )

        return inner_box


@dataclass
class DoorLikeFactory(SemanticAnnotationFactory[T], ABC):
    """
    Abstract factory for creating door-like factories such as doors or double doors.
    """


@dataclass
class DoorFactory(DoorLikeFactory[Door], HasHandleFactory):
    """
    Factory for creating a door with a handle. The door is defined by its scale and handle direction.
    The doors origin is at the pivot point of the door, not at the center.
    """

    scale: Scale = field(default_factory=lambda: Scale(0.03, 1.0, 2.0))
    """
    The scale of the entryway.
    """

    def _create(self, world: World) -> World:
        """
        Return a world with a door body at its root. The door has a handle and is defined by its scale and handle direction.
        """

        door_event = self.scale.simple_event.as_composite_set()

        body = Body(name=self.name)
        bounding_box_collection = BoundingBoxCollection.from_event(body, door_event)
        collision = bounding_box_collection.as_shapes()
        body.collision = collision
        body.visual = collision

        world.add_kinematic_structure_entity(body)

        door_T_handle = (
            self.parent_T_handle
            or self.create_parent_T_handle_from_parent_scale(self.scale)
        )
        self.add_handle_to_world(door_T_handle, world)
        semantic_handle_annotation: Handle = world.get_semantic_annotations_by_type(
            Handle
        )[0]
        world.add_semantic_annotation(
            Door(name=self.name, handle=semantic_handle_annotation, body=body)
        )

        return world


@dataclass
class DoubleDoorFactory(DoorLikeFactory[DoubleDoor], HasDoorLikeFactories):
    """
    Factory for creating a double door with two doors and their handles.
    """

    def __post_init__(self):
        assert (
            len(self.door_factories) == len(self.door_transforms) == 2
        ), "Double door must have exactly two door factories and transforms"

    def _create(self, world: World) -> World:
        """
        Return a world with a virtual body at its root that is the parent of the two doors making up the double door.
        """

        double_door_body = Body(name=self.name)
        world.add_kinematic_structure_entity(double_door_body)

        self.add_doorlike_semantic_annotation_to_world(
            parent_world=world,
        )

        semantic_door_annotations = world.get_semantic_annotations_by_type(Door)
        assert (
            len(semantic_door_annotations) == 2
        ), "Double door must have exactly two doors semantic annotations"

        left_door, right_door = semantic_door_annotations
        if (
            left_door.body.parent_connection.origin_expression.y
            > right_door.body.parent_connection.origin_expression.y
        ):
            right_door, left_door = (
                semantic_door_annotations[0],
                semantic_door_annotations[1],
            )

        semantic_double_door_annotation = DoubleDoor(
            body=double_door_body, left_door=left_door, right_door=right_door
        )
        world.add_semantic_annotation(semantic_double_door_annotation)
        return world


@dataclass
class DrawerFactory(SemanticAnnotationFactory[Drawer], HasHandleFactory):
    """
    Factory for creating a drawer with a handle and a container.
    """

    container_factory: ContainerFactory = field(default=None)
    """
    The factory used to create the container of the drawer.
    """

    def _create(self, world: World) -> World:
        """
        Return a world with a drawer at its root. The drawer consists of a container and a handle.
        """

        container_world = self.container_factory.create()
        world.merge_world(container_world)
        parent_T_handle = (
            self.parent_T_handle
            or self.create_parent_T_handle_from_parent_scale(
                self.container_factory.scale
            )
        )

        self.add_handle_to_world(parent_T_handle, world)

        semantic_container_annotation: Container = (
            world.get_semantic_annotations_by_type(Container)[0]
        )
        semantic_handle_annotation: Handle = world.get_semantic_annotations_by_type(
            Handle
        )[0]
        semantic_drawer_annotation = Drawer(
            name=self.name,
            container=semantic_container_annotation,
            handle=semantic_handle_annotation,
        )
        world.add_semantic_annotation(semantic_drawer_annotation)

        return world


@dataclass
class DresserFactory(
    SemanticAnnotationFactory[Dresser], HasDoorLikeFactories, HasDrawerFactories
):
    """
    Factory for creating a dresser with drawers, and doors.
    """

    container_factory: ContainerFactory = field(default=None)
    """
    The factory used to create the container of the dresser.
    """

    def _create(self, world: World) -> World:
        """
        Return a world with a dresser at its root. The dresser consists of a container, potentially drawers, and doors.
        Assumes that the number of drawers matches the number of drawer transforms.
        """
        assert len(self.drawers_factories) == len(
            self.parent_T_drawers
        ), "Number of drawers must match number of transforms"

        dresser_world = self._make_dresser_world()
        dresser_world.name = world.name
        return self._make_interior(dresser_world)

    def _make_dresser_world(self) -> World:
        """
        Create a world with a dresser semantic annotation that contains a container, drawers, and doors, but no interior yet.
        """
        dresser_world = self.container_factory.create()
        with dresser_world.modify_world():
            semantic_container_annotation: Container = (
                dresser_world.get_semantic_annotations_by_type(Container)[0]
            )

            self.add_doorlike_semantic_annotation_to_world(dresser_world)

            self.add_drawers_to_world(dresser_world)

            semantic_dresser_annotation = Dresser(
                name=self.name,
                container=semantic_container_annotation,
                drawers=dresser_world.get_semantic_annotations_by_type(Drawer),
                doors=dresser_world.get_semantic_annotations_by_type(Door),
            )
            dresser_world.add_semantic_annotation(
                semantic_dresser_annotation, exists_ok=True
            )
            dresser_world.name = self.name.name

        return dresser_world

    def _make_interior(self, world: World) -> World:
        """
        Create the interior of the dresser by subtracting the drawers and doors from the container, and filling  with
        the remaining space.

        :param world: The world containing the dresser body as its root.
        """
        dresser_body: Body = world.root
        container_event = dresser_body.collision.as_bounding_box_collection_at_origin(
            TransformationMatrix(reference_frame=dresser_body)
        ).event

        container_footprint = self._subtract_bodies_from_container_footprint(
            world, container_event
        )

        container_event = self._fill_container_body(
            container_footprint, container_event
        )

        collision_shapes = BoundingBoxCollection.from_event(
            dresser_body, container_event
        ).as_shapes()
        dresser_body.collision = collision_shapes
        dresser_body.visual = collision_shapes
        return world

    def _subtract_bodies_from_container_footprint(
        self, world: World, container_event: Event
    ) -> Event:
        """
        Subtract the bounding boxes of all bodies in the world from the container event,
        except for the dresser body itself. This creates a frontal footprint of the container

        :param world: The world containing the dresser body as its root.
        :param container_event: The event representing the container.

        :return: An event representing the footprint of the container after subtracting other bodies.
        """
        dresser_body = world.root

        container_footprint = container_event.marginal(SpatialVariables.yz)

        for body in world.bodies_with_enabled_collision:
            if body == dresser_body:
                continue
            body_footprint = body.collision.as_bounding_box_collection_at_origin(
                TransformationMatrix(reference_frame=dresser_body)
            ).event.marginal(SpatialVariables.yz)
            container_footprint -= body_footprint
            if container_footprint.is_empty():
                return Event()

        return container_footprint

    def _fill_container_body(
        self, container_footprint: Event, container_event: Event
    ) -> Event:
        """
        Expand container footprint into 3d space and fill the space of the resulting container body.

        :param container_footprint: The footprint of the container in the yz-plane.
        :param container_event: The event representing the container.

        :return: An event representing the container body with the footprint filled in the x-direction.
        """

        container_footprint.fill_missing_variables([SpatialVariables.x.value])

        depth_interval = container_event.bounding_box()[SpatialVariables.x.value]
        limiting_event = SimpleEvent(
            {SpatialVariables.x.value: depth_interval}
        ).as_composite_set()
        limiting_event.fill_missing_variables(SpatialVariables.yz)

        container_event |= container_footprint & limiting_event

        return container_event


@dataclass
class RoomFactory(SemanticAnnotationFactory[Room]):
    """
    Factory for creating a room with a specific region.
    """

    floor_polytope: List[Point3]
    """
    The region that defines the room's boundaries and reference frame.
    """

    def _create(self, world: World) -> World:
        """
        Return a world with a room semantic annotation that contains the specified region.
        """
        room_body = Body(name=self.name)
        world.add_kinematic_structure_entity(room_body)

        region = Region.from_3d_points(
            points_3d=self.floor_polytope,
            name=PrefixedName(self.name.name + "_region", self.name.prefix),
            reference_frame=room_body,
        )
        connection = FixedConnection(
            parent=room_body,
            child=region,
            parent_T_connection_expression=TransformationMatrix(),
        )
        world.add_connection(connection)

        floor = Floor(
            name=PrefixedName(self.name.name + "_floor", self.name.prefix),
            region=region,
        )
        world.add_semantic_annotation(floor)
        semantic_room_annotation = Room(name=self.name, floor=floor)
        world.add_semantic_annotation(semantic_room_annotation)

        return world


@dataclass
class WallFactory(SemanticAnnotationFactory[Wall], HasDoorLikeFactories):

    scale: Scale = field(kw_only=True)
    """
    The scale of the wall.
    """

    def _create(self, world: World) -> World:
        """
        Return a world with the wall body at its root and potentially doors and double doors as children of the wall body.
        """
        wall_world = self._create_wall_world()
        self.add_doorlike_semantic_annotation_to_world(wall_world)
        self.remove_doors_from_world(wall_world)
        world.merge_world(wall_world)

        return world

    def _create_wall_world(self) -> World:
        wall_world = World()
        with wall_world.modify_world():
            wall_body = Body(name=self.name)
            wall_collision = self._create_wall_collision(wall_body)
            wall_body.collision = wall_collision
            wall_body.visual = wall_collision
            with wall_world.modify_world():
                wall_world.add_kinematic_structure_entity(wall_body)

            wall = Wall(
                name=self.name,
                body=wall_body,
            )

            wall_world.add_semantic_annotation(wall)

        return wall_world

    def _create_wall_collision(self, reference_frame: Body) -> ShapeCollection:
        """
        Return the collision shapes for the wall. A wall event is created based on the scale of the wall, and
        doors are removed from the wall event. The resulting bounding box collection is converted to shapes.
        """

        wall_event = self._create_wall_event().as_composite_set()

        bounding_box_collection = BoundingBoxCollection.from_event(
            reference_frame, wall_event
        )

        wall_collision = bounding_box_collection.as_shapes()
        return wall_collision

    def _create_wall_event(self) -> SimpleEvent:
        """
        Return a wall event created from its scale. The height origin is on the ground, not in the center of the wall.
        """
        x_interval = closed(-self.scale.x / 2, self.scale.x / 2)
        y_interval = closed(-self.scale.y / 2, self.scale.y / 2)
        z_interval = closed(0, self.scale.z)

        wall_event = SimpleEvent(
            {
                SpatialVariables.x.value: x_interval,
                SpatialVariables.y.value: y_interval,
                SpatialVariables.z.value: z_interval,
            }
        )
        return wall_event



from __future__ import annotations

from enum import Enum
from functools import cached_property

from krrood.ormatic.utils import classproperty
from random_events.variable import Continuous
from sortedcontainers import SortedSet


class SpatialVariables(Enum):
    """
    Enum for spatial variables used in the semantic digital twin. Used in the context of random events.
    """

    x = Continuous("x")
    y = Continuous("y")
    z = Continuous("z")

    @classproperty
    def xy(cls):
        return SortedSet([cls.x.value, cls.y.value])

    @classproperty
    def xz(cls):
        return SortedSet([cls.x.value, cls.z.value])

    @classproperty
    def yz(cls):
        return SortedSet([cls.y.value, cls.z.value])



from abc import ABC
from dataclasses import dataclass
from functools import lru_cache
from typing_extensions import Dict, Any, Self, List

from random_events.utils import SubclassJSONSerializer

from ...datastructures.prefixed_name import PrefixedName
from ...world_description.world_modification import (
    WorldModelModificationBlock,
)


@dataclass
class MetaData(SubclassJSONSerializer):
    """
    Class for data describing the origin of a message.
    """

    node_name: str
    """
    The name of the node that published this message
    """

    process_id: int
    """
    The id of the process that published this message
    """

    object_id: int
    """
    The id of the object in the process that issues this publishing call
    """

    @lru_cache(maxsize=None)
    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "node_name": self.node_name,
            "process_id": self.process_id,
            "object_id": self.object_id,
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            node_name=data["node_name"],
            process_id=data["process_id"],
            object_id=data["object_id"],
        )

    def __hash__(self):
        return hash((self.node_name, self.process_id, self.object_id))


@dataclass
class Message(SubclassJSONSerializer, ABC):

    meta_data: MetaData
    """
    Message origin meta data.
    """

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "meta_data": self.meta_data.to_json(),
        }


@dataclass
class WorldStateUpdate(Message):
    """
    Class describing the updates to the free variables of a world state.
    """

    prefixed_names: List[PrefixedName]
    """
    The prefixed names of the changed free variables.
    """

    states: List[float]
    """
    The states of the changed free variables.
    """

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "prefixed_names": [n.to_json() for n in self.prefixed_names],
            "states": list(self.states),
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            meta_data=MetaData.from_json(data["meta_data"]),
            prefixed_names=[PrefixedName.from_json(n) for n in data["prefixed_names"]],
            states=data["states"],
        )


@dataclass
class ModificationBlock(Message):
    """
    Message describing the modifications done to a world.
    """

    modifications: WorldModelModificationBlock
    """
    The modifications done to a world.
    """

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "modifications": self.modifications.to_json(),
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            meta_data=MetaData.from_json(data["meta_data"]),
            modifications=WorldModelModificationBlock.from_json(data["modifications"]),
        )


@dataclass
class LoadModel(Message):
    """
    Message for requesting the loading of a model identified by its primary key.
    """

    primary_key: int
    """
    The primary key identifying the model to be loaded.
    """

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "primary_key": self.primary_key,
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            meta_data=MetaData.from_json(data["meta_data"]),
            primary_key=data["primary_key"],
        )



from copy import deepcopy

import numpy as np
import pytest
from semantic_digital_twin.semantic_annotations.semantic_annotations import Handle

from semantic_digital_twin.spatial_types import Vector3
from semantic_digital_twin.world import World
from semantic_digital_twin.world_description.connections import (
    PrismaticConnection,
    RevoluteConnection,
    Connection6DoF,
    FixedConnection,
)
from semantic_digital_twin.exceptions import (
    AddingAnExistingSemanticAnnotationError,
    DuplicateSemanticAnnotationError,
    SemanticAnnotationNotFoundError,
    DuplicateKinematicStructureEntityError,
    UsageError,
    MissingWorldModificationContextError,
)
from semantic_digital_twin.datastructures.prefixed_name import PrefixedName
from semantic_digital_twin.spatial_types.derivatives import Derivatives, DerivativeMap

# from semantic_digital_twin.spatial_types.math import rotation_matrix_from_rpy
from semantic_digital_twin.spatial_types.spatial_types import (
    TransformationMatrix,
    RotationMatrix,
)
from semantic_digital_twin.spatial_types.symbol_manager import symbol_manager
from semantic_digital_twin.testing import world_setup, pr2_world
from semantic_digital_twin.world_description.world_entity import (
    SemanticAnnotation,
    Body,
)


def test_set_state(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    c1: PrismaticConnection = world.get_connection(l1, l2)
    c1.position = 1.0
    assert c1.position == 1.0
    c2: RevoluteConnection = world.get_connection(r1, r2)
    c2.position = 1337
    assert c2.position == 1337
    c3: Connection6DoF = world.get_connection(world.root, bf)
    transform = RotationMatrix.from_rpy(1, 0, 0).to_np()
    transform[0, 3] = 69
    c3.origin = transform
    assert np.allclose(world.compute_forward_kinematics_np(world.root, bf), transform)

    world.set_positions_1DOF_connection({c1: 2})
    assert c1.position == 2.0

    transform[0, 3] += c1.position
    assert np.allclose(l2.global_pose.to_np(), transform)


def test_construction(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    world.validate()
    assert len(world.connections) == 5
    assert len(world.kinematic_structure_entities) == 6
    assert world.state.positions[0] == 0
    assert (
        world.get_connection(l1, l2).dof.name == world.get_connection(r1, r2).dof.name
    )


def test_chain_of_bodies(world_setup):
    world, _, l2, _, _, _ = world_setup
    result = world.compute_chain_of_kinematic_structure_entities(
        root=world.root, tip=l2
    )
    result = [x.name for x in result]
    assert result == [
        PrefixedName(name="root", prefix="world"),
        PrefixedName(name="bf", prefix=None),
        PrefixedName(name="l1", prefix=None),
        PrefixedName(name="l2", prefix=None),
    ]


def test_chain_of_connections(world_setup):
    world, _, l2, _, _, _ = world_setup
    result = world.compute_chain_of_connections(root=world.root, tip=l2)
    result = [x.name for x in result]
    assert result == [
        PrefixedName(name="root_T_bf", prefix=None),
        PrefixedName(name="bf_T_l1", prefix=None),
        PrefixedName(name="l1_T_l2", prefix=None),
    ]


def test_split_chain_of_bodies(world_setup):
    world, _, l2, _, _, r2 = world_setup
    result = world.compute_split_chain_of_kinematic_structure_entities(root=r2, tip=l2)
    result = tuple([x.name for x in y] for y in result)
    assert result == (
        [PrefixedName(name="r2", prefix=None), PrefixedName(name="r1", prefix=None)],
        [PrefixedName(name="bf", prefix=None)],
        [PrefixedName(name="l1", prefix=None), PrefixedName(name="l2", prefix=None)],
    )


def test_split_chain_of_bodies_adjacent1(world_setup):
    world, _, _, _, r1, r2 = world_setup
    result = world.compute_split_chain_of_kinematic_structure_entities(root=r2, tip=r1)
    result = tuple([x.name for x in y] for y in result)
    assert result == (
        [PrefixedName(name="r2", prefix=None)],
        [PrefixedName(name="r1", prefix=None)],
        [],
    )


def test_split_chain_of_bodies_adjacent2(world_setup):
    world, _, _, _, r1, r2 = world_setup
    result = world.compute_split_chain_of_kinematic_structure_entities(root=r1, tip=r2)
    result = tuple([x.name for x in y] for y in result)
    assert result == (
        [],
        [PrefixedName(name="r1", prefix=None)],
        [PrefixedName(name="r2", prefix=None)],
    )


def test_split_chain_of_bodies_identical(world_setup):
    world, _, _, _, r1, _ = world_setup
    result = world.compute_split_chain_of_kinematic_structure_entities(root=r1, tip=r1)
    result = tuple([x.name for x in y] for y in result)
    assert result == ([], [PrefixedName(name="r1", prefix=None)], [])


def test_split_chain_of_connections(world_setup):
    world, _, l2, _, _, r2 = world_setup
    result = world.compute_split_chain_of_connections(root=r2, tip=l2)
    result = tuple([x.name for x in y] for y in result)
    assert result == (
        [
            PrefixedName(name="r1_T_r2", prefix=None),
            PrefixedName(name="bf_T_r1", prefix=None),
        ],
        [
            PrefixedName(name="bf_T_l1", prefix=None),
            PrefixedName(name="l1_T_l2", prefix=None),
        ],
    )


def test_split_chain_of_connections_adjacent1(world_setup):
    world, _, _, _, r1, r2 = world_setup
    result = world.compute_split_chain_of_connections(root=r2, tip=r1)
    result = tuple([x.name for x in y] for y in result)
    assert result == ([PrefixedName(name="r1_T_r2", prefix=None)], [])


def test_split_chain_of_connections_adjacent2(world_setup):
    world, _, _, _, r1, r2 = world_setup
    result = world.compute_split_chain_of_connections(root=r1, tip=r2)
    result = tuple([x.name for x in y] for y in result)
    assert result == ([], [PrefixedName(name="r1_T_r2", prefix=None)])


def test_split_chain_of_connections_identical(world_setup):
    world, _, _, _, r1, _ = world_setup
    result = world.compute_split_chain_of_connections(root=r1, tip=r1)
    result = tuple([x.name for x in y] for y in result)
    assert result == ([], [])


@pytest.mark.skip(
    reason="readding of 1dof connection broken because reference to dof is lost"
)
def test_nested_with_blocks_illegal_state(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup

    with world.modify_world():
        connection1 = world.get_connection(l1, l2)
        world.remove_connection(connection1)
        with world.modify_world():
            connection2 = world.get_connection(r1, r2)
            world.remove_connection(connection2)
        world.add_connection(connection1)
        world.add_connection(connection2)


def test_compute_fk_connection6dof(world_setup):
    world, _, _, bf, _, _ = world_setup
    fk = world.compute_forward_kinematics_np(world.root, bf)
    np.testing.assert_array_equal(fk, np.eye(4))

    connection: Connection6DoF = world.get_connection(world.root, bf)

    world.state[connection.x.name].position = 1.0
    world.state[connection.qw.name].position = 0
    world.state[connection.qz.name].position = 1
    world.notify_state_change()
    fk = world.compute_forward_kinematics_np(world.root, bf)
    np.testing.assert_array_equal(
        fk,
        [
            [-1.0, 0.0, 0.0, 1.0],
            [0.0, -1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ],
    )


def test_compute_fk(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    fk = world.compute_forward_kinematics_np(l2, r2)
    np.testing.assert_array_equal(fk, np.eye(4))

    connection: PrismaticConnection = world.get_connection(r1, r2)

    world.state[connection.dof.name].position = 1.0
    world.notify_state_change()
    fk = world.compute_forward_kinematics_np(l2, r2)
    assert np.allclose(
        fk,
        np.array(
            [
                [0.540302, -0.841471, 0.0, -1.0],
                [0.841471, 0.540302, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        ),
    )


def test_compute_ik(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    target = np.array(
        [
            [0.540302, -0.841471, 0.0, -1.0],
            [0.841471, 0.540302, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    joint_state = world.compute_inverse_kinematics(
        l2, r2, TransformationMatrix(target, reference_frame=l2)
    )
    for joint, state in joint_state.items():
        world.state[joint.name].position = state
    world.notify_state_change()
    assert np.allclose(world.compute_forward_kinematics_np(l2, r2), target, atol=1e-3)


def test_compute_fk_expression(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    connection: PrismaticConnection = world.get_connection(r1, r2)
    world.state[connection.dof.name].position = 1.0
    world.notify_state_change()
    fk = world.compute_forward_kinematics_np(r2, l2)
    fk_expr = world.compose_forward_kinematics_expression(r2, l2)
    fk_expr_compiled = fk_expr.compile()
    fk2 = fk_expr_compiled(
        *symbol_manager.resolve_symbols(fk_expr_compiled.symbol_parameters)
    )
    np.testing.assert_array_almost_equal(fk, fk2)


def test_apply_control_commands(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    connection: PrismaticConnection = world.get_connection(r1, r2)
    cmd = np.array([100.0, 0, 0, 0, 0, 0, 0, 0])
    dt = 0.1
    world.apply_control_commands(cmd, dt, Derivatives.jerk)
    assert world.state[connection.dof.name].jerk == 100.0
    assert world.state[connection.dof.name].acceleration == 100.0 * dt
    assert world.state[connection.dof.name].velocity == 100.0 * dt * dt
    assert world.state[connection.dof.name].position == 100.0 * dt * dt * dt


def test_compute_relative_pose(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    connection: PrismaticConnection = world.get_connection(l1, l2)
    world.state[connection.dof.name].position = 1.0
    world.notify_state_change()

    pose = TransformationMatrix(reference_frame=l2)
    relative_pose = world.transform(pose, l1)
    expected_pose = TransformationMatrix(
        [
            [1.0, 0, 0.0, 1.0],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    np.testing.assert_array_almost_equal(relative_pose.to_np(), expected_pose.to_np())


def test_compute_relative_pose_both(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    world.get_connection(world.root, bf).origin = np.array(
        [
            [0.0, -1.0, 0.0, 1.0],
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    world.notify_state_change()

    pose = TransformationMatrix.from_xyz_rpy(x=1.0, reference_frame=bf)
    relative_pose = world.transform(pose, world.root)
    # Rotation is 90 degrees around z-axis, translation is 1 along x-axis
    expected_pose = np.array(
        [
            [0.0, -1.0, 0.0, 1.0],
            [1.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    np.testing.assert_array_almost_equal(relative_pose.to_np(), expected_pose)


def test_compute_relative_pose_only_translation(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    connection: PrismaticConnection = world.get_connection(l1, l2)
    world.state[connection.dof.name].position = 1.0
    world.notify_state_change()

    pose = TransformationMatrix.from_xyz_rpy(x=2.0, reference_frame=l2)
    relative_pose = world.transform(pose, l1)
    expected_pose = np.array(
        [
            [1.0, 0, 0.0, 3.0],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    np.testing.assert_array_almost_equal(relative_pose.to_np(), expected_pose)


def test_compute_relative_pose_only_rotation(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    connection: RevoluteConnection = world.get_connection(r1, r2)
    world.state[connection.dof.name].position = np.pi / 2  # 90 degrees
    world.notify_state_change()

    pose = TransformationMatrix(reference_frame=r2)
    relative_pose = world.transform(pose, r1)
    expected_pose = np.array(
        [
            [0.0, -1.0, 0.0, 0.0],
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    np.testing.assert_array_almost_equal(relative_pose.to_np(), expected_pose)


def test_add_semantic_annotation(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    v = SemanticAnnotation(name=PrefixedName("muh"))
    with world.modify_world():
        world.add_semantic_annotation(v)
    with pytest.raises(AddingAnExistingSemanticAnnotationError):
        world.add_semantic_annotation(v, exists_ok=False)
    assert world.get_semantic_annotation_by_name(v.name) == v


def test_duplicate_semantic_annotation(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    v = SemanticAnnotation(name=PrefixedName("muh"))
    with world.modify_world():
        world.add_semantic_annotation(v)
        world.semantic_annotations.append(v)
    with pytest.raises(DuplicateSemanticAnnotationError):
        world.get_semantic_annotation_by_name(v.name)


def test_merge_world(world_setup, pr2_world):
    world, l1, l2, bf, r1, r2 = world_setup

    base_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_link")
    )
    r_gripper_tool_frame = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    torso_lift_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("torso_lift_link")
    )
    r_shoulder_pan_joint = pr2_world.get_connection(
        torso_lift_link,
        pr2_world.get_kinematic_structure_entity_by_name(
            PrefixedName("r_shoulder_pan_link")
        ),
    )

    l_shoulder_pan_joint = pr2_world.get_connection(
        torso_lift_link,
        pr2_world.get_kinematic_structure_entity_by_name(
            PrefixedName("l_shoulder_pan_link")
        ),
    )

    world.merge_world(pr2_world)

    assert base_link in world.kinematic_structure_entities
    assert r_gripper_tool_frame in world.kinematic_structure_entities
    assert l_shoulder_pan_joint in world.connections
    assert torso_lift_link._world == world
    assert r_shoulder_pan_joint._world == world


def test_merge_with_connection(world_setup, pr2_world):
    world, l1, l2, bf, r1, r2 = world_setup

    base_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_link")
    )
    r_gripper_tool_frame = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    torso_lift_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("torso_lift_link")
    )
    r_shoulder_pan_joint = pr2_world.get_connection(
        torso_lift_link,
        pr2_world.get_kinematic_structure_entity_by_name(
            PrefixedName("r_shoulder_pan_link")
        ),
    )

    pose = np.eye(4)
    pose[0, 3] = 1.0

    origin = TransformationMatrix(pose)

    connection = pr2_world.get_connection_by_name("l_gripper_l_finger_joint")
    pr2_world.state[connection.dof.name].position = 0.55
    pr2_world.notify_state_change()
    expected_fk = pr2_world.compute_forward_kinematics(
        connection.parent, connection.child
    ).to_np()

    new_connection = FixedConnection(
        parent=world.root,
        child=pr2_world.root,
        parent_T_connection_expression=origin,
    )

    world.merge_world(pr2_world, new_connection)
    assert base_link in world.kinematic_structure_entities
    assert r_gripper_tool_frame in world.kinematic_structure_entities
    assert new_connection in world.connections
    assert torso_lift_link._world == world
    assert r_shoulder_pan_joint._world == world
    assert world.state[connection.dof.name].position == 0.55
    assert world.compute_forward_kinematics_np(world.root, base_link)[
        0, 3
    ] == pytest.approx(1.0, abs=1e-6)
    actual_fk = world.compute_forward_kinematics(
        connection.parent, connection.child
    ).to_np()
    assert np.allclose(actual_fk, expected_fk)


def test_merge_with_pose(world_setup, pr2_world):
    world, l1, l2, bf, r1, r2 = world_setup

    base_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_link")
    )
    r_gripper_tool_frame = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    torso_lift_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("torso_lift_link")
    )
    r_shoulder_pan_joint = pr2_world.get_connection(
        torso_lift_link,
        pr2_world.get_kinematic_structure_entity_by_name(
            PrefixedName("r_shoulder_pan_link")
        ),
    )

    pose = np.eye(4)
    pose[0, 3] = 1.0  # Translate along x-axis

    world.merge_world_at_pose(pr2_world, TransformationMatrix(pose))

    assert base_link in world.kinematic_structure_entities
    assert r_gripper_tool_frame in world.kinematic_structure_entities
    assert torso_lift_link._world == world
    assert r_shoulder_pan_joint._world == world
    assert world.compute_forward_kinematics_np(world.root, base_link)[
        0, 3
    ] == pytest.approx(1.0, abs=1e-6)


def test_merge_with_pose_rotation(world_setup, pr2_world):
    world, l1, l2, bf, r1, r2 = world_setup

    base_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_link")
    )
    r_gripper_tool_frame = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    torso_lift_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("torso_lift_link")
    )
    r_shoulder_pan_joint = pr2_world.get_connection(
        torso_lift_link,
        pr2_world.get_kinematic_structure_entity_by_name(
            PrefixedName("r_shoulder_pan_link")
        ),
    )
    base_footprint = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_footprint")
    )

    # Rotation is 90 degrees around z-axis, translation is 1 along x-axis
    pose = np.array(
        [
            [0.0, -1.0, 0.0, 1.0],
            [1.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    world.merge_world_at_pose(pr2_world, TransformationMatrix(pose))

    assert base_link in world.kinematic_structure_entities
    assert r_gripper_tool_frame in world.kinematic_structure_entities
    assert torso_lift_link._world == world
    assert r_shoulder_pan_joint._world == world
    fk_base = world.compute_forward_kinematics_np(world.root, base_footprint)
    assert fk_base[0, 3] == pytest.approx(1.0, abs=1e-6)
    assert fk_base[1, 3] == pytest.approx(1.0, abs=1e-6)
    assert fk_base[2, 3] == pytest.approx(0.0, abs=1e-6)
    np.testing.assert_array_almost_equal(
        RotationMatrix.from_rpy(0, 0, np.pi / 2).to_np()[:3, :3],
        fk_base[:3, :3],
        decimal=6,
    )


def test_merge_in_empty_world(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup

    empty_world = World()

    assert empty_world.root is None
    empty_world.merge_world(world)
    assert empty_world.root is not None


def test_remove_connection(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    connection = world.get_connection(l1, l2)
    num_dofs = len(world.degrees_of_freedom)
    with world.modify_world():
        world.remove_connection(connection)
        world.remove_kinematic_structure_entity(l2)
    assert connection not in world.connections
    # dof should still exist because it was a mimic connection, so the number didn't change.
    assert num_dofs == len(world.degrees_of_freedom)

    with world.modify_world():
        world.remove_connection(world.get_connection(r1, r2))
        new_connection = FixedConnection(r1, r2)
        world.add_connection(new_connection)

    with pytest.raises(ValueError):
        # if you remove a connection, the child must be connected some other way or deleted
        world.remove_connection(world.get_connection(r1, r2))


def test_copy_world(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    world_copy = deepcopy(world)
    assert l2 not in world_copy.bodies
    assert bf.parent_connection not in world_copy.connections
    bf.parent_connection.origin = np.array(
        [[1, 0, 0, 1.5], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    )
    assert (
        float(
            world_copy.get_kinematic_structure_entity_by_name("bf").global_pose.to_np()[
                0, 3
            ]
        )
        == 0.0
    )
    assert float(bf.global_pose.to_np()[0, 3]) == 1.5


def test_copy_world_state(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    connection: PrismaticConnection = world.get_connection(r1, r2)
    world.state[connection.dof.name].position = 1.0
    world.notify_state_change()
    world_copy = deepcopy(world)

    assert world.get_connection(r1, r2).position == 1.0
    assert world_copy.get_connection(r1, r2).position == 1.0


def test_match_index(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    world_copy = deepcopy(world)
    for body in world.bodies:
        new_body = world_copy.get_kinematic_structure_entity_by_name(body.name)
        assert body.index == new_body.index


def test_copy_dof(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    world_copy = deepcopy(world)
    for dof in world.degrees_of_freedom:
        new_dof = world_copy.get_degree_of_freedom_by_name(dof.name)
        assert dof.name == new_dof.name
        assert dof.lower_limits == new_dof.lower_limits
        assert dof.upper_limits == new_dof.upper_limits


def test_copy_pr2_world(pr2_world):
    pr2_world.state[
        pr2_world.get_degree_of_freedom_by_name("torso_lift_joint").name
    ].position = 0.3
    pr2_world.notify_state_change()
    pr2_copy = deepcopy(pr2_world)


def test_copy_pr2_world_connection_origin(pr2_world):
    pr2_world.notify_state_change()
    pr2_copy = deepcopy(pr2_world)

    for body in pr2_world.bodies:
        pr2_body = pr2_world.get_kinematic_structure_entity_by_name(body.name)
        pr2_copy_body = pr2_copy.get_kinematic_structure_entity_by_name(body.name)
        np.testing.assert_array_almost_equal(
            pr2_body.global_pose.to_np(), pr2_copy_body.global_pose.to_np()
        )


def test_world_different_entities(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    world_copy = deepcopy(world)
    for body in world_copy.bodies:
        assert body not in world.bodies
    for connection in world_copy.connections:
        assert connection not in world.connections
    for dof in world_copy.state:
        assert dof not in world.degrees_of_freedom


def test_copy_pr2(pr2_world):
    pr2_world.state[
        pr2_world.get_degree_of_freedom_by_name("torso_lift_joint").name
    ].position = 0.3
    pr2_world.notify_state_change()
    pr2_copy = deepcopy(pr2_world)
    assert pr2_world.get_kinematic_structure_entity_by_name(
        "head_tilt_link"
    ).global_pose.to_np()[2, 3] == pytest.approx(1.472, abs=1e-3)
    assert pr2_copy.get_kinematic_structure_entity_by_name(
        "head_tilt_link"
    ).global_pose.to_np()[2, 3] == pytest.approx(1.472, abs=1e-3)


def test_add_entity_with_duplicate_name(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    body_duplicate = Body(name=PrefixedName("l1"))
    with pytest.raises(DuplicateKinematicStructureEntityError):
        with world.modify_world():
            world.add_kinematic_structure_entity(body_duplicate)


def test_overwrite_dof_limits(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    connection: PrismaticConnection = world.get_connections_by_type(
        PrismaticConnection
    )[0]
    assert connection.dof.lower_limits.velocity == -1
    assert connection.dof.upper_limits.velocity == 1

    new_limits = DerivativeMap([0.69, 0.42, 1337, 23])

    connection.raw_dof._overwrite_dof_limits(
        new_lower_limits=new_limits * -1, new_upper_limits=new_limits
    )
    assert connection.dof.lower_limits.position == -new_limits.position
    assert connection.dof.upper_limits.position == new_limits.position
    assert connection.dof.lower_limits.velocity == -new_limits.velocity
    assert connection.dof.upper_limits.velocity == new_limits.velocity
    assert connection.dof.lower_limits.acceleration == -new_limits.acceleration
    assert connection.dof.upper_limits.acceleration == new_limits.acceleration
    assert connection.dof.lower_limits.jerk == -new_limits.jerk
    assert connection.dof.upper_limits.jerk == new_limits.jerk

    new_limits2 = DerivativeMap([3333, 3333, 3333, 3333])

    connection.raw_dof._overwrite_dof_limits(
        new_lower_limits=new_limits2 * -1, new_upper_limits=new_limits2
    )
    assert connection.dof.lower_limits.position == -new_limits.position
    assert connection.dof.upper_limits.position == new_limits.position
    assert connection.dof.lower_limits.velocity == -new_limits.velocity
    assert connection.dof.upper_limits.velocity == new_limits.velocity
    assert connection.dof.lower_limits.acceleration == -new_limits.acceleration
    assert connection.dof.upper_limits.acceleration == new_limits.acceleration
    assert connection.dof.lower_limits.jerk == -new_limits.jerk
    assert connection.dof.upper_limits.jerk == new_limits.jerk


def test_overwrite_dof_limits_mimic(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    connection: PrismaticConnection = world.get_connections_by_type(
        PrismaticConnection
    )[0]
    with world.modify_world():
        body = Body(name=PrefixedName("muh"))
        mimic_connection = PrismaticConnection(
            parent=bf,
            child=body,
            offset=23,
            multiplier=-2,
            axis=Vector3(0, 0, 1),
            dof_name=connection.dof_name,
        )
        world.add_body(body)
        world.add_connection(mimic_connection)

    # when the multiplier is negative, the vel limits shouldn't be swapped
    assert np.isclose(
        mimic_connection.dof.lower_limits.velocity,
        connection.dof.lower_limits.velocity * 2,
    )
    assert np.isclose(
        mimic_connection.dof.upper_limits.velocity,
        connection.dof.upper_limits.velocity * 2,
    )

    new_limits = DerivativeMap([0.69, 0.42, 1337, 23])

    with pytest.raises(UsageError):
        mimic_connection.dof._overwrite_dof_limits(
            new_lower_limits=new_limits * -1, new_upper_limits=new_limits
        )

    mimic_connection.raw_dof._overwrite_dof_limits(
        new_lower_limits=new_limits * -1, new_upper_limits=new_limits
    )

    # Check that limits are correctly applied with negative multiplier and offset
    # Position limits: swapped due to negative multiplier, then scaled and offset applied
    # Lower becomes: new_limits.position * (-2) + 23 = 0.69 * (-2) + 23 = -1.38 + 23 = 21.62
    # Upper becomes: (new_limits * -1).position * (-2) + 23 = -0.69 * (-2) + 23 = 1.38 + 23 = 24.38
    assert np.isclose(mimic_connection.dof.lower_limits.position, 21.62)
    assert np.isclose(mimic_connection.dof.upper_limits.position, 24.38)

    # Velocity limits: only multiplier applied (no offset), but absolute value for limits
    # Since we're dealing with limits, velocity should use abs(multiplier) = 2
    assert np.isclose(
        mimic_connection.dof.lower_limits.velocity, (new_limits * -1).velocity * 2
    )
    assert np.isclose(
        mimic_connection.dof.upper_limits.velocity, new_limits.velocity * 2
    )

    # Acceleration limits: only multiplier applied (no offset), absolute value for limits
    assert np.isclose(
        mimic_connection.dof.lower_limits.acceleration,
        (new_limits * -1).acceleration * 2,
    )
    assert np.isclose(
        mimic_connection.dof.upper_limits.acceleration, new_limits.acceleration * 2
    )

    # Jerk limits: only multiplier applied (no offset), absolute value for limits
    assert np.isclose(
        mimic_connection.dof.lower_limits.jerk, (new_limits * -1).jerk * 2
    )
    assert np.isclose(mimic_connection.dof.upper_limits.jerk, new_limits.jerk * 2)

    # limits are only applied if the new ones are lower
    new_limits2 = DerivativeMap([3333, 3333, 3333, 3333])

    mimic_connection.raw_dof._overwrite_dof_limits(
        new_lower_limits=new_limits2 * -1, new_upper_limits=new_limits2
    )

    assert np.isclose(mimic_connection.dof.lower_limits.position, 21.62)
    assert np.isclose(mimic_connection.dof.upper_limits.position, 24.38)

    assert np.isclose(
        mimic_connection.dof.lower_limits.velocity, (new_limits * -1).velocity * 2
    )
    assert np.isclose(
        mimic_connection.dof.upper_limits.velocity, new_limits.velocity * 2
    )

    assert np.isclose(
        mimic_connection.dof.lower_limits.acceleration,
        (new_limits * -1).acceleration * 2,
    )
    assert np.isclose(
        mimic_connection.dof.upper_limits.acceleration, new_limits.acceleration * 2
    )

    assert np.isclose(
        mimic_connection.dof.lower_limits.jerk, (new_limits * -1).jerk * 2
    )
    assert np.isclose(mimic_connection.dof.upper_limits.jerk, new_limits.jerk * 2)


def test_missing_world_modification_context(world_setup):
    world, l1, l2, bf, r1, r2 = world_setup
    with pytest.raises(MissingWorldModificationContextError):
        world.add_semantic_annotation(Handle(l1))



from ...world import World
from typing_extensions import Any, Dict
from ripple_down_rules.datastructures.case import Case, create_case
from ripple_down_rules.helpers import general_rdr_classify
from . import world_semantic_annotations_mcrdr as semantic_annotations_classifier

name = "world"
case_type = World
case_name = "World"
classifiers_dict = dict()
classifiers_dict["semantic_annotations"] = semantic_annotations_classifier


def classify(case: World, **kwargs) -> Dict[str, Any]:
    if not isinstance(case, Case):
        case = create_case(case, max_recursion_idx=3)
    return general_rdr_classify(classifiers_dict, case, **kwargs)
